/**
 * Compile with: mex -v predict.cpp ShapeRFClassifier.cpp -I../../boost_1_54_0
 *
 * This code accompanies the paper:
 *
 * "Dense non-rigid shape correspondence using random forests"
 * E.Rodola, S.Rota Bulo, T.Windheuser, M.Vestner, D.Cremers.
 * Proc. CVPR 2014
 *
 * Please cite the paper above if you use the code.
 *
 * Emanuele Rodolà (c) 2014
 */
#include "mex.h"
#include <iostream>
#include "ShapeRFClassifier.h"
#include "mesh.h"

#include <fstream>

void mexFunction(int nlhs, mxArray *plhs[], int nrhs, const mxArray*prhs[])
{
	if (nrhs != 6 || nlhs != 3)
      mexErrMsgTxt("Usage: [r,c,v] = predict(shape, trees, n_eigenvectors, n_energy_levels, max_likelihood, n_tot_labels).");

	const double* const trees = mxGetPr(prhs[1]);
	const int n_trees = int( mxGetN(prhs[1]) );
	
	const int n_eigenvectors = static_cast<int>(*mxGetPr(prhs[2]));
	const int n_energy_levels = static_cast<int>(*mxGetPr(prhs[3]));
	
	const int n_tot_labels = static_cast<int>(*mxGetPr(prhs[5]));
	
	const bool max_likelihood = (*mxGetPr(prhs[4]) == 0. ? false : true);
	
	const std::string test_shape = mxArrayToString(prhs[0]);
	
	if (n_trees < 1)
		mexErrMsgTxt("The forest should have at least one tree.");
		
	if (n_eigenvectors < 2)
		mexErrMsgTxt("At least 2 eigenvectors are required.");
		
	if (test_shape.empty())
		mexErrMsgTxt("The test shape must be a valid path.");
	
	std::vector<std::string> tree_fnames(n_trees);

	for (int i=0; i<n_trees; ++i)
	{
		mxArray* fld = mxGetFieldByNumber(prhs[1], 0, i);
		if (fld==0)
			mexErrMsgTxt("Error when parsing the trees struct.");
		else
			tree_fnames[i] = std::string(mxArrayToString(fld));
	}
	
	//DEBUG
	//std::ofstream fout("./debug.txt", std::ios::out);
	
	try
	{
		int np;
		std::vector<double> test_evals(n_eigenvectors);
		{
			const std::string fname = test_shape + ".evals.dat";
			std::ifstream fin(fname.c_str(), std::ios::in);
			
			if (!fin.is_open())
			{
				const std::string msg = "Cannot open " + fname;
				mexErrMsgTxt(msg.c_str());
			}
				
			double tmp;
			fin >> tmp;
			np = static_cast<int>(tmp);
			
			for (int i=0; i<n_eigenvectors; ++i)
				fin >> test_evals[i];
			
			fin.close();
		}
		
		std::vector<double> test_evecs(np*n_eigenvectors);
		{
			const std::string fname = test_shape + ".evecs.dat";
			std::ifstream fin(fname.c_str(), std::ios::in);
			
			if (!fin.is_open())
			{
				const std::string msg = "Cannot open " + fname;
				mexErrMsgTxt(msg.c_str());
			}
			
			for (size_t i=0; i<test_evecs.size(); ++i)
				fin >> test_evecs[i];
			
			fin.close();
		}
		
		DataSet testing_dataset(n_eigenvectors, n_energy_levels);
		testing_dataset.addShape(std::vector<int>(np,0), std::vector<int>(), test_evals, test_evecs);
		
		std::vector<ShapePrediction> predictions(np);
		
		for (int cur_tree=0; cur_tree<n_trees; ++cur_tree)
		{
			ShapeRandomTree tree;
			const std::string& fname = tree_fnames[cur_tree];
			
			std::ifstream fin(fname.c_str(), std::ios::in);
			if (!fin.is_open())
			{
				const std::string msg = "Cannot open " + fname;
				mexErrMsgTxt(msg.c_str());
			}
			else
				fin.close();

			tree.load_binary(fname);

			tree.setDataSet(&testing_dataset);

			for (int pt=0; pt<np; ++pt)
			{
				ShapePrediction& prediction = predictions[pt];

				if (cur_tree==0)
					prediction = tree.predict(pt);
				else
					prediction.add( tree.predict(pt) );
			}
		} // next tree
		
		std::vector< std::pair< std::pair<int,int> , double > > sparse_mat;
		sparse_mat.reserve(np);
		
		if (max_likelihood)
		{
			for (int pt=0; pt<np; ++pt)
			{
				ShapePrediction& prediction = predictions[pt];
				prediction.normalize();
				sparse_mat.push_back( std::make_pair( std::make_pair(pt+1, prediction.bestLabel()) , 1.0 ) );
			}
		}
		else
		{
			for (int pt=0; pt<np; ++pt)
			{
				ShapePrediction& prediction = predictions[pt];
				prediction.normalize();
				for (std::map<uint32_t,float>::iterator it=prediction.prob.begin(); it!=prediction.prob.end(); ++it)
					//out_labels[pt*n_tot_labels+it->first-1] = it->second;
					sparse_mat.push_back( std::make_pair( std::make_pair(pt+1, it->first) , it->second ) );
			}
		}
		
		plhs[0] = mxCreateDoubleMatrix(sparse_mat.size(), 1, mxREAL);
		plhs[1] = mxCreateDoubleMatrix(sparse_mat.size(), 1, mxREAL);
		plhs[2] = mxCreateDoubleMatrix(sparse_mat.size(), 1, mxREAL);
		double* out_rows = mxGetPr(plhs[0]);
		double* out_cols = mxGetPr(plhs[1]);
		double* out_vals = mxGetPr(plhs[2]);
		
		for (int i=0; i<sparse_mat.size(); ++i)
		{
			out_rows[i] = sparse_mat[i].first.first;
			out_cols[i] = sparse_mat[i].first.second;
			out_vals[i] = sparse_mat[i].second;
		}
	}
	catch (std::exception& e)
	{
		const std::string msg = "Exception caught: " + std::string(e.what());
		mexErrMsgTxt(msg.c_str());
	}
}
